package cc.fascinated.services;

import cc.fascinated.exception.impl.BadRequestException;
import cc.fascinated.model.leaderboard.Leaderboard;
import cc.fascinated.model.score.ScoreResponse;
import cc.fascinated.model.score.ScoresOverResponse;
import cc.fascinated.model.score.TotalScoresResponse;
import cc.fascinated.model.score.TrackedScore;
import cc.fascinated.model.user.User;
import cc.fascinated.platform.Platform;
import cc.fascinated.repository.couchdb.TrackedScoreRepository;
import lombok.NonNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.LinkedList;
import java.util.List;

/**
 * @author Fascinated (fascinated7)
 */
@Service
public class TrackedScoreService {
    /**
     * The scores over thresholds.
     */
    private static final int[] SCORES_OVER = {1000, 900, 800, 700, 600, 500, 400, 300, 200, 100};

    /**
     * The tracked score repository to use.
     */
    @NonNull
    private final TrackedScoreRepository trackedScoreRepository;

    /**
     * The user service to use.
     */
    @NonNull
    private final UserService userService;

    /**
     * The ScoreSaber service to use.
     */
    @NonNull
    private final ScoreSaberService scoreSaberService;

    @Autowired
    public TrackedScoreService(@NonNull TrackedScoreRepository trackedScoreRepository, @NonNull UserService userService,
                               @NonNull ScoreSaberService scoreSaberService) {
        this.trackedScoreRepository = trackedScoreRepository;
        this.userService = userService;
        this.scoreSaberService = scoreSaberService;
        this.trackedScoreRepository.ensureDeduplication();
    }

    /**
     * Gets a list of top tracked scores
     * sorted by pp from the platform
     *
     * @param platform the platform to get the scores from
     * @param amount   the amount of scores to get
     * @return the scores
     */
    public List<ScoreResponse> getTopScores(Platform.Platforms platform, int amount) {
        List<TrackedScore> foundScores = trackedScoreRepository.findTopRankedScores(platform.getPlatformName(), amount);
        if (foundScores.isEmpty()) {
            throw new BadRequestException("No scores found for platform " + platform.getPlatformName());
        }

        List<ScoreResponse> scores = new LinkedList<>();
        for (TrackedScore trackedScore : foundScores) {
            User user = this.userService.getUser(trackedScore.getPlayerId());
            Leaderboard leaderboard = null;
            switch (platform) {
                case SCORESABER -> leaderboard = Leaderboard.getFromScoreSaberToken(this.scoreSaberService.getLeaderboard(trackedScore.getLeaderboardId()));
            }
            assert leaderboard != null; // This should never be null

            scores.add(new ScoreResponse(
                    trackedScore.getScoreId(),
                    trackedScore.getPp(),
                    trackedScore.getRank(),
                    trackedScore.getScore(),
                    trackedScore.getModifiedScore(),
                    trackedScore.getWeight(),
                    trackedScore.getModifiers(),
                    trackedScore.getMultiplier(),
                    trackedScore.getMissedNotes(),
                    trackedScore.getBadCuts(),
                    trackedScore.getMaxCombo(),
                    trackedScore.getAccuracy(),
                    user,
                    leaderboard,
                    trackedScore.getTimestamp()
            ));
        }

        return scores;
    }

    /**
     * Gets the amount of scores over pp thresholds.
     *
     * @param platform the platform to get the scores from
     * @return the scores over pp thresholds
     */
    public ScoresOverResponse getScoresOver(Platform.Platforms platform) {
        ScoresOverResponse scoresOverResponse = new ScoresOverResponse();

        for (int i : SCORES_OVER) {
            scoresOverResponse.addScores(i, trackedScoreRepository.getScoreCountOverPpThreshold(platform.getPlatformName(), i));
        }
        return scoresOverResponse;
    }

    /**
     * Gets the total amount of scores
     * for a platform.
     *
     * @param platform the platform to get the scores from
     * @return the total amount of scores for the platform
     */
    public TotalScoresResponse getTotalScores(Platform.Platforms platform) {
        return new TotalScoresResponse(
                trackedScoreRepository.countTotalScores(platform.getPlatformName()),
                trackedScoreRepository.countTotalRankedScores(platform.getPlatformName())
        );
    }

    /**
     * Gets a list of tracked scores
     * for a platform.
     *
     * @param platform the platform to get the scores from
     * @return the tracked scores
     */
    public List<TrackedScore> getTrackedScores(Platform.Platforms platform, boolean onlyRanked) {
        if (onlyRanked) {
            return trackedScoreRepository.findAllByPlatformRankedOnly(platform.getPlatformName());
        }
        return trackedScoreRepository.findAllByPlatform(platform.getPlatformName());
    }

    /**
     * Saves a list of tracked scores.
     *
     * @param scores the scores to save
     */
    public void updateScores(TrackedScore... scores) {
        for (TrackedScore score : scores) {
            this.trackedScoreRepository.updateScorePp(score.getScoreId(), score.getPp());
        }
    }

    /**
     * Deletes a list of tracked scores.
     *
     * @param scores the scores to delete
     */
    public void deleteScores(TrackedScore... scores) {
        for (TrackedScore score : scores) {
            this.trackedScoreRepository.delete(score);
        }
    }
}
